;#pragma usePoFile 0
; Helper scripts for the Teamtalk application.
;
; Author:  Doug Lee
/*
Copyright (c) 2010-2018 Doug Lee.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* The names of the copyright holders and contributors may not be used to
  endorse or promote products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

include "hjconst.jsh"
include "tt_jcpdict.jsm"

;========================== Module jcp (from jcpdict) ==========================
; Control property manager for JAWS using collections.
; Callback: jcp__setCustomProps.
; Set all control-specific properties there to make them work in speech and Braille consistently.
;
; Author:  Doug Lee

include "winstyles.jsh"
const
; Default cache lifetime for jcp__setAll() in milliseconds.
	jcp___DefaultCacheTicks = 500

globals
; Master on/off switch for the system, set by jcp__enable(), used for testing.
	int jcp___disabled,
; Cache lifetime in effect at any given time.
	int jcp___cacheTicks,
; True when this system is already active (used to avoid "unknown function call" messages)
	int jcp___inside,
; Last script key and name of script that was just run.
; Set by KeyPressedEvent (via jcp___setCaller) and used by jcp__getActionClass().
	string jcp___key,
	string jcp___caller,
; Caches for getObjectSubtypeCode() and jcp__getID(),
; because they can get called below enough times per second to slow things down.
; Example symptom: A one-second delay on a PCCursor() script call.
	collection jcp___intcache,
	int jcp___intTick

variant function jcp___vcast(variant v)
return v
endFunction

void function autoStartEvent()
jcp___cacheTicks = jcp___DefaultCacheTicks
jcp___intcache = new collection
endFunction

int function jcp__enable(int flag)
; Master on/off switch:  pass 0 to turn off, 1 for on, 2 to query without changing, and anything else to toggle.
; (This arrangement was chosen to make direct on/off make sense while making it easy to use isSameScript() to query on one press and toggle on more.)
if flag == 0 || flag == 1
	jcp___disabled = !flag
elif flag != 2
	jcp___disabled = !jcp___disabled
endIf
if jcp___disabled
	sayMessage(OT_No_Disable, "JCP disabled")
else
	sayMessage(OT_No_Disable, "JCP enabled")
endIf
return !jcp___disabled
endFunction

int function jcp__cache(int ticks)
; Set how long the cache of control properties can be kept before a forced refresh.
; Determines how often setCustomControlProps() is called.
; Pass 0 to use the internal default value.
; WARNING: Setting this too low can cause a major performance hit on systems with Braille displays,
; because setCustomControlProps() can be called over 40 times per second in such cases.
; Returns the old value that is being replaced.
; ticks: The maximum number of ticks (milliseconds) to keep the cache valid,
; or 0 to use the internal default value.
var
	int old
old = jcp___cacheTicks
if !ticks
	ticks = jcp___DefaultCacheTicks
endIf
jcp___cacheTicks = ticks
return old
endFunction

int function jcp__isEmpty(collection props)
; Returns True if props represents an empty property set.
var string k
forEach k in props
	; If this line is reached, there's at least one property.
	return False
endForEach
; And if this line is reached, there are no properties.
return True
endFunction

collection function jcp__new()
; Returns an empty props structure.
var collection c c = new collection
return c
endFunction

string function jcp__getActionClass()
; Get the type of action that appears to have caused the current jcpDict invocations:
;	tab:  Navigation among controls.
;	arrow:  Navigation within a control.
var
	string id
id = stringLower(jcp___key)
if stringContains(id, "tab")
	return "tab"
elif stringContains(id, "arrow")
|| stringContains(id, "home")
|| stringContains(id, "end")
	return "arrow"
endIf
return ""
endFunction

variant function jcp__get(collection props, string whichProp)
; Get the value of whichProp from props and return it.
; Case is not significant in property names.
; Dotted subclassing like fieldname.brl is permitted:
; If fieldname.brl exists it is returned, else if fieldname exists it is returned, else null is returned.
; Implemented subclassings:
;	- .brl for Braille calls.
;	- .<caller> for functions in this file that call jcp__get().
;	- .<class> for keys/scripts that initiated this jcp__get call (see jcp__getActionClass()).
; The latter of these is implemented internally here, not passed in whichProp.
; <caller> takes precedence over <class>.
; Function-call values are supported safely; see jcp__set().
var
	int pos,
	string val,
	string class,
	int done
val = ""
done = False
class = ""
while !done
	val = ""
	if collectionItemExists(props, whichProp)
		val = props[whichProp]
	endIf
	if val
		if props[whichProp+".isFunc"]
			val = formatStringWithEmbeddedFunctions("<" +val +">")
		endIf
		return val
	endIf
	pos = stringContainsFromRight(whichProp, ".")
	if pos
		whichProp = stringLeft(whichProp, pos-1)
		if !whichProp
			done = True
		endIf
	elif !class
		class = jcp__getActionClass()
		if class
			whichProp = whichProp +"." +jcp__getActionClass()
		else
			done = True
		endIf
	else
		done = True
	endIf
endWhile
return ""
endFunction

int function jcp__prepend(collection props, string whichProp, string s)
; Prepend s to the property value indicated.
; whichProp should indicate a string property, not a numeric one.
var string val = jcp__get(props, whichProp)
if val
	val = s +" " +val
else
	val = s
endIf
jcp__set(props, whichProp, val)
endFunction

int function jcp__append(collection props, string whichProp, string s)
; Append s to the property value indicated.
; whichProp should indicate a string property, not a numeric one.
var string val = jcp__get(props, whichProp)
if val
	val = val +" " +s
else
	val = s
endIf
jcp__set(props, whichProp, val)
endFunction

void function jcp__setBits(collection props, string whichProp, int bits)
; Logically OR a bit or set of bits into a property value.
; whichProp should indicate a numeric property, not a string one.
var int val = jcp__get(props, whichProp) +0
val = val | bits
jcp__set(props, whichProp, val)
endFunction

void function jcp__clearBits(collection props, string whichProp, int bits)
; Logically AND a bit or set of bits out of a property value.
; whichProp should indicate a numeric property, not a string one.
var int val = jcp__get(props, whichProp) +0
val = val & (~bits)
jcp__set(props, whichProp, val)
endFunction

variant function jcp__setItemPos(collection props, variant pos, variant cnt)
; Set the Position property by its component parts.
; Also sets ItemPos and ItemCount.
; Recommended over setting Position or ItemPos/ItemCount directly.
; For convenience, stringToInt conversions are performed here if necessary
; so callers don't have to do it.
var int p = stringToInt(pos)
var int c = stringToInt(cnt)
props.itemPos = p
props.itemCount = c
var string buf
if p
	buf = formatString(jcp___MOfN, intToString(p), intToString(c))
else
	buf = formatString(jcp___NItems, intToString(c))
endIf
jcp__set(props, "Position", buf)
return True
endFunction

int function jcp__setTablePos(collection props, int rowIdx, int rowCount, int colIdx, int colCount)
; Set table row/column position and count properties.
; The "+0"'s avoid beeps when nulls are passed in.
return jcp__set(props, "row", rowIdx+0)
&& jcp__set(props, "rows", rowCount+0)
&& jcp__set(props, "col", colIdx+0)
&& jcp__set(props, "cols", colCount+0)
endFunction

string function jcp__getID(string sOrigin)
var
	int left, int right, int top, int bottom
if sOrigin == "current" || (sOrigin == "focus" && isPCCursor())
	if getObjectRect(left, right, top, bottom)
		return "obj" +intToString(left) +"," +intToString(top)
	else
		; Use the active cursor location.
		left = getCursorCol()
		top = getCursorRow()
		return "point" +intToString(left) +"," +intToString(top)
	endIf
elif sOrigin == "focus"
	; Focus but PC cursor not active; use the PC cursor location explicitly.
	saveCursor()  PCCursor()
	left = getCursorCol()
	top = getCursorRow()
	restoreCursor()
	return "point" +intToString(left) +"," +intToString(top)
else
	; sOrigin is not focus or current.  Use sOrigin itself as an ID.
	return sOrigin
endIf
endFunction

globals
	int jcp___lastTick,
	string jcp___lastID,
	string jcp___lastProps
int function jcp__setAll(collection byRef props, string sOrigin, /*?*/ int forceNoCache)
; Sets props to be a structure for other controlprop* functions.
; Obtains any properties of the indicated control that stray from what JAWS would normally discover and report.
; Returns True if anything was done or null if no special handling of this control is necessary.
var
	string id,
	int tc
; Start with a clean slate.
props = jcp__new()
if userBufferIsActive()
	return False
endIf
if !sOrigin
	sOrigin = "focus"
endIf

; Cache property sets (avoids serious CPU hits when a Braille display is active)
tc = getTickCount()
id = jcp___cachedVal(sOrigin)
if !forceNoCache
&& jcp___lastTick
&& tc -jcp___lastTick <= jcp___cacheTicks
&& id && !stringCompare(jcp___lastID, id, True)
	props = jcp___vcast(jcp___lastProps)
	return !jcp__isEmpty(props)
endIf
jcp___lastID = jcp___vcast(id)
jcp___lastTick = tc

jcp___lastProps = jcp___vcast(props)

; Set any custom properties via the callback function.
;var int tcj = getTickCount()
jcp__setCustomProps(props, sOrigin)
;sayString(formatString("%2", id, intToString(getTickCount()-tcj)))
if !jcp__isEmpty(props)
	; Set the origin so jcp__say() will know where to get defaults.
	if !stringLength(jcp__get(props, "origin"))
		jcp__set(props, "origin", sOrigin)
	endIf
	jcp___lastProps = jcp___vcast(props)
	return True
endIf
return False
endFunction

int function jcp__setByOrigin(string sOrigin, string request, handle byRef hwnd, object byRef o, int byRef childID)
; Set origin info by reference.
; Note: This function is deprecated as of 2015-04-22; please use jcp__getLocation().
; Allowed values for request:
;	h - Window handle (NOTE: w in jcp__getLocation()).
;	m - MSAA object and childID
;	n - Nav ID (implies h).
;	c - Cursor-centric function support (e.g., getObjectName etc.).
; If passing c, first activate the cursor that will be used for getObject* etc. calls.
; At present, m and n are mutually exclusive.
; Returns True on success and False on failure.
request = stringLower(request)

; hwnd origin case.
if stringLeft(sOrigin, 4) == "hwnd"
	hwnd = stringToHandle(stringChopLeft(sOrigin, 4))
	; Hwnd comes back whether requested or not here.
	if stringContains(request, "m")
		o = getObjectFromEvent(hwnd, -4, 0, childID)
	elif stringContains(request, "n")
		; ToDo: Probably insufficient.
		childID = jcp___vcast(navGetCurrentObjectID())
	endIf
	if stringContains(request, "c")
		if getCurrentWindow() != hwnd
			moveToWindow(hwnd)
			; TODO:  Verify that this returns the right window.
		endIf
	endIf
	return True
endIf

; Focus or current-when-current-is-focus cases.
if sOrigin == "focus" || (sOrigin == "current" && isPCCursor())
	hwnd = getFocus()
	if stringContains(request, "m")
		o = getFocusObject(childID)
		if !o
			return False
		endIf
	elif stringContains(request, "n")
		childID = jcp___vcast(navGetFocusObjectID())
	endIf
	if !isPCCursor()
		; Make GetObject*() calls apply to the right thing.
		saveCursor()  PCCursor()
	endIf
	return True
endIf

; Current (when distinct from focus) case.
if sOrigin == "current"
	hwnd = getCurrentWindow()
	if stringContains(request, "m")
		o = getObjectFromEvent(hwnd, -4, 0, childID)
		if !o
			return False
		endIf
	elif stringContains(request, "n")
		childID = jcp___vcast(navGetCurrentObjectID())
	endIf
	return True
endIf

outputDebugString(formatString("jcp__setCustomProps: Unsupported origin: %1", sOrigin))
return False
endFunction

collection function jcp__getLocation(string sOrigin, string request)
; Return as a collection information about the location indicated by sOrigin.
; Possible values of sOrigin:
;	focus: The focused location.
;	current: The current location (same as focus when the PC cursor is active).
;	subtype<code>: Nearest parent location from current with given subtypeCode  (for BrailleAddObject* functions).
;	hwnd<handle>: A specific window by its handle.
; See jcp__setCustomProps() for further information about sOrigin.
; Allowed values for request:
;	w - Window handle.
;	m - MSAA object and childID
;	n - Nav ID (implies w).
;	u - UIA element (ToDo: Not yet supported to avoid loading a lot of code when not needed).
;	d - DOM object (ToDo: Not supported and probably destined to go away along with IE itself).
;	c - Cursor-centric function support (e.g., getObjectName etc.).
; If c is passed, the currently active cursor can be changed and/or moved.
; Returns null on failure.
; Fields in the returned collection and which request values make them appear:
;	origin: The value of sOrigin on entry (but may be adjusted; e.g., current+PCCursor = focus) (all).
;	request: The value of request on entry (all).
;	reply: The request flags that could actually be honored (all).
;	hwnd: The window handle (w and others).
;	MSAA.obj, MSAA.childID: Collection of IAccessible object and id (m).
;	navID: The nav*() function module navigation ID (n).
;	UIAElement: The UIA element (u).
;	DOMElement: The DOM element (d).
var collection info info = new collection
info.origin = sOrigin
request = stringLower(request)
info.request = request
info.reply = request
var collection MSAA
if stringContains(request, "m")
	MSAA = new collection
	info.MSAA = MSAA
endIf
var int childID

; hwnd origin case.
if stringLeft(sOrigin, 4) == "hwnd"
	info.hwnd = stringToHandle(stringChopLeft(sOrigin, 4))
	; Hwnd comes back whether requested or not.
	if stringContains(request, "m")
		MSAA.obj = getObjectFromEvent(info.hwnd, -4, 0, childID)
		MSAA.childID = childID
		if !MSAA.obj
			info.reply = stringReplaceSubstrings(info.reply, "m", "")
		endIf
	endIf
	if stringContains(request, "n")
		; ToDo: Probably insufficient.
		info.navID = jcp___vcast(navGetCurrentObjectID())
		if !info.navID
			info.reply = stringReplaceSubstrings(info.reply, "n", "")
		endIf
	endIf
	if stringContains(request, "c")
		; Invisible cursor used here; can clobber current position if already active.
		if getCurrentWindow() != info.hwnd
			if !isInvisibleCursor()
				saveCursor() invisibleCursor()
			endIf
			moveToWindow(info.hwnd)
			; TODO:  Verify that this returns the right window.
		endIf
	endIf
	return info
endIf

; Focus or current-when-current-is-focus cases.
if sOrigin == "focus" || (sOrigin == "current" && isPCCursor())
	if info.origin != "focus"
		info.origin = "focus"
	endIf
	info.hwnd = getFocus()
	if stringContains(request, "m")
		MSAA.obj = getFocusObject(childID)
		MSAA.childID = childID
		if MSAA.obj
			; This may find a different childID than MSAA.childID.
			; This has been known to help (e.g., Skype).
			var variant v = MSAA.obj.accFocus
			if stringToInt(v)
				MSAA.childID = v
			elif v
				MSAA.obj = v
				MSAA.childID = 0
				v = MSAA.obj.accSelection
				if stringToInt(v)
					MSAA.childID = v
				; Can't just test v in case it's an array of results.
				; Every valid MSAA object in an app should have an accParent,
				; but a selection array will not.
				elif v.accParent
					MSAA.obj = v
					MSAA.childID = 0
				endIf
			endIf
		endIf
		if !MSAA.obj
			info.reply = stringReplaceSubstrings(info.reply, "m", "")
		endIf
	endIf
	if stringContains(request, "n")
		info.navID = jcp___vcast(navGetFocusObjectID())
		if !info.navID
			info.reply = stringReplaceSubstrings(info.reply, "n", "")
		endIf
	endIf
	if stringContains(request, "c") && !isPCCursor()
		; Make GetObject*() calls apply to the right thing.
		PCCursor()
	endIf
	return info
endIf

; Current (when distinct from focus) case.
if sOrigin == "current"
	info.hwnd = getCurrentWindow()
	if stringContains(request, "m")
		MSAA.obj = getObjectFromEvent(info.hwnd, -4, 0, childID)
		MSAA.childID = childID
		if !MSAA.obj
			info.reply = stringReplaceSubstrings(info.reply, "m", "")
		endIf
	endIf
	if stringContains(request, "n")
		info.navID = jcp___vcast(navGetCurrentObjectID())
		if !info.navID
			info.reply = stringReplaceSubstrings(info.reply, "n", "")
		endIf
	endIf
	return info
endIf

outputDebugString(formatString("jcp__setCustomProps: Unsupported origin: %1", sOrigin))
return
endFunction

void function jcp__reset()
; Force the next jcp__setAll to reload, not use the cache.
jcp___lastTick = 0
jcp___intTick = 0
if BrailleInUse()
	scheduleFunction("jcp___BrailleRefresh", 2)
endIf
endFunction

void function jcp___BrailleRefresh()
; Refresh the Braille display.
BrailleRefresh(brlShowSameRange)
endFunction

/*
void function jcp__setCustomProps(collection props, string sOrigin)
; Set any custom properties required per situation.
; Callers are expected to override this function.
; props is empty on entry and can be modified but should not be assigned to.
; Properties understood by default (property name case insignificant):
; Properties used by both Braille and speech:
;	Name, type, state
;	containerName, containerType
;	value, position, dlgText
;	level (for treeViews)
;	itemPos, itemCount (convenient way to compose position)
;	Hotkey (used in SayTutorialHelpHotKey)
;	subtypeCode (default type for speech and SayTutorialHelp and returned by BrailleCallbackObjectIdentify for this control)
;	attributes (CTRL_* constants from hjconst.jsh defining the control's states)
;	beforeField, afterField (text to speak/Braille before/after this field)
; Speech-only properties:
;	tutor, beforeTutor (tutor overrides the normal tutor message for the subtype)
;	customTutor: Becomes a custom tutor message that can speak when standard ones are turned off by the user.
; Braille-only properties:
;	dlgPageName, contextHelp, time
; Special properties internal to this system:
;	Mask:  Masks all special handling (use with caller name, e.g., jcp__set(props, "Mask.SayLine", True)).
;	Origin:  The origin to use to get default property values with getObject* functions (usually the sOrigin value used to create this property set)
; Note that SubtypeCode will set Type if Type is not set already; same for Attributes and State.
; sOrigin values that must be handled:
;	focus, current
;	subtype<subtype> (for BrailleAddObject* functions)
;	hwnd<handle>
; Other possible sOrigin values:
;	point<x>,<y>:  Like current but at a specific point
;	obj[<level>]  (level 0 is current)
;	hwnd<handle>/{nav<id>|sdm<id>|MSAA<id>}:  Nav* function ID, SDM ID, or MSAA ID within a window
;	point<x>,<y>:{hwnd|nav|SDM|MSAA}:  Like the extended hwnd syntax but from a specific point
endFunction
*/

variant function jcp___cachedVal(string which)
var int tc = getTickCount()
if jcp___intTick && tc -jcp___intTick < 250
	if collectionItemExists(jcp___intcache, which)
		return jcp___intcache[which]
	endIf
else
	collectionRemoveAll(jcp___intcache)
	jcp___intTick = tc
endIf
; The sought value does not exist now.
if which == "stc"
	jcp___intcache[which] = getObjectSubtypeCode()
else
	; An id request.
	jcp___intcache[which] = jcp__getID(which)
endIf
return jcp___intcache[which]
endFunction

int function jcp___sayControl(handle hwnd, string name, string type, string state, string containerName, string containerType, string value, string position, string dialogText)
var
	string allAttribs,
	int markupIncluded
allAttribs = name +type +state +containerName +containerType +value +position +dialogText
if !allAttribs
	return False
endIf
markupIncluded = (stringContains(stringLower(allAttribs), "<voice") > 0)
if !markupIncluded
	name = SMMReplaceSymbolsWithMarkup(name)
	type = SMMReplaceSymbolsWithMarkup(type)
	state = SMMReplaceSymbolsWithMarkup(state)
	containerName = SMMReplaceSymbolsWithMarkup(containerName)
	containerType = SMMReplaceSymbolsWithMarkup(containerType)
	value = SMMReplaceSymbolsWithMarkup(value)
	position = SMMReplaceSymbolsWithMarkup(position)
	dialogText = SMMReplaceSymbolsWithMarkup(dialogText)
endIf
if dialogText
	sayControlExWithMarkup(hwnd, name, type, state, containerName, containerType, value, position, dialogText)
elif position
	sayControlExWithMarkup(hwnd, name, type, state, containerName, containerType, value, position)
elif value
	sayControlExWithMarkup(hwnd, name, type, state, containerName, containerType, value)
elif containerType
	sayControlExWithMarkup(hwnd, name, type, state, containerName, containerType)
elif containerName
	sayControlExWithMarkup(hwnd, name, type, state, containerName)
elif state
	sayControlExWithMarkup(hwnd, name, type, state)
elif type
	sayControlExWithMarkup(hwnd, name, type)
else
	sayControlExWithMarkup(hwnd, name)
endIf
return True
endFunction

variant function jcp___dispatch(string func)
return formatStringWithEmbeddedFunctions("<" +func +">")
endFunction

int function jcp__say(collection props, string caller)
; Returns True if processing was completed here and False if it should still be done elsewhere.
var
	string origin,
	string callerName,
	string actionClass,
	string beforeField, string afterField,
	string name, string type, string state, string containerName, string containerType, string value, string position, string dialogText,
	int rowIdx, int rowCount, int colIdx, int colCount,
	int typeCode, int attributes,
	handle hwnd
callerName = stringSegment(caller, "(", 1)
var int spelling = False
if callerName == "spellLine"
	callerName = "sayLine"
	spelling = True
elif caller == "SayByTypeForScriptSayLine"
	callerName = "sayLine"
endIf
if jcp__get(props, "Mask."+callerName)
	return False
endIf
origin = jcp__get(props, "origin")
if origin == "focus"
	hwnd = getFocus()
elif origin == "current"
	hwnd = getCurrentWindow()
elif stringLeft(origin, 4) == "hwnd"
	hwnd = stringToHandle(stringChopLeft(origin, 4))
else
	beep()
	return False
endIf

; Because of how they're used, null and unset string values are treated equal here.
; More care must be taken with numeric values though.
name = jcp__get(props, "name."+callerName)
type = jcp__get(props, "type."+callerName)
if !type
	typeCode = jcp__get(props, "subtypeCode."+callerName)
	; This one is treated the same if 0 or unset.
	if typeCode
		type = smmStripMarkup(smmGetStartMarkupForControlType(typeCode))
		if !type
			type = " "
		endIf
	endIf
endIf
state = jcp__get(props, "state."+callerName)
if !state
	attributes = jcp__get(props, "attributes."+callerName)
	; Attributes of 0 are supported explicitly here.
	if collectionItemExists(props, "attributes")
		if !typeCode
			typeCode = 1  ; all seem to work equally well
		endIf
		state = smmStripMarkup(smmGetStartMarkupForControlState(typeCode, attributes))
		if !state
			state = " "
		endIf
		if !type
			type = " "
		endIf
	endIf
endIf
containerName = jcp__get(props, "containerName."+callerName)
containerType = jcp__get(props, "containerType."+callerName)
value = jcp__get(props, "value."+callerName)
position = jcp__get(props, "position."+callerName)
dialogText = jcp__get(props, "DlgText."+callerName)
; beforeField is useful when assigning a name adversely changes other elements of how JAWS says a control.
beforeField = jcp__get(props, "beforeField."+callerName)
; afterField is useful for things like "minutes" in "Mark me away after _____ minutes"
afterField = jcp__get(props, "afterField."+callerName)

if stringLeft(origin, 4) == "hwnd"
	saveCursor() invisibleCursor() saveCursor()
	moveToWindow(hwnd)
endIf

actionClass = jcp__getActionClass()
	if caller != "sayLine" && actionClass != "arrow" && beforeField
		say(beforeField, OT_Control_Name, True)
	endIf

if caller == "sayLine" || actionClass == "arrow"
	containerName = ""
	containerType = ""
	dialogText = ""
endIf

if stringLength(name +type +state +containerName +containerType +value +position +dialogText)
	; Use SayControlEx or SayControlExWithMarkup.
	var int customName, int customType, int customState, int customValue, int customPosition
	customName = stringLength(name)
	if !name name = getObjectName() endIf
	customType = stringLength(type)
	if !type
		; Convert types that JAWS normally does not announce directly.
		var int tcode = getObjectSubtypeCode()
		var int tcode1 = 0
		if tcode == WT_ListItem && WT_ListItem == 86
			; WT_List == 85 but some non-English JAWS versions are
			; missing this constant in hjconst.jsh.
			tcode1 = 85
		elif tcode == WT_ListBoxItem
			tcode1 = WT_ListBox
		elif tcode == WT_ListViewItem
			tcode1 = WT_ListView
		elif tcode == WT_TreeViewItem
			tcode1 = WT_TreeView
		endIf
		if tcode1
			type = smmStripMarkup(smmGetStartMarkupForControlType(tcode1))
		else
			type = getObjectSubtype()
		endIf
	endIf
	customState = stringLength(state)
	if !state
		state = getObjectState()
		; JAWS does not normally say "selected," so get rid of it when present here.
		state = stringTrimLeadingBlanks(stringTrimTrailingBlanks(stringReplaceSubstrings(state, jcp___stateSelected, "")))
		state = stringReplaceSubstrings(state, "  ", " ")
	endIf
	; No function for containerName/Type.
	customValue = stringLength(value)
	if !value value = getObjectValue() endIf
	customPosition = stringLength(position)
	if !position position = positionInGroup() endIf
	; TODO: No default dialog text here, could try getDialogStaticText().

	; Caller-specific adjustments.
	var int isVirtual = isVirtualPCCursor()
	if isVirtual
	&& stringContains(":statictext:graphic:", ":"+stringLower(stringStripAllBlanks(type))+":")
		type = " "
	endIf
	if caller == "sayLine" || actionClass == "arrow"
		containerName = " "
		containerType = " "
		dialogText = " "
		if !isVirtual
		&& !stringContains(":button:radiobutton:edit:readonlyedit:", ":"+stringLower(stringStripAllBlanks(type))+":")
			type = " "
		elif isVirtual
		&& stringContains(":statictext:graphic:", ":"+stringLower(stringStripAllBlanks(type))+":")
			type = " "
		endIf
		if !stringIsBlank(value)
			name = " "
		endIf
		if caller != "sayLine"
		&& !isVirtual
			position = " "
		endIf
	elif stringStartsWith(caller, "sayObjectActiveItem")
		beforeField = ""
		type = " "
		containerName = " "
		dialogText = " "
		if !stringIsBlank(value)
			name = " "
		endIf
		if stringContains(caller, "0")
			; Position info is not to be spoken.
			position = " "
		endIf
	else
		; Special for tab controls to avoid Tab <name> Tab.
		if tcode == WT_TabControl || tcode1 == WT_TabControl
			containerType = " "
		endIf
	endIf

	if spelling
		spellString(name +value)
	elif !jcp___sayControl(hwnd, name, type, state, containerName, containerType, value, position, dialogText)
		jcp___dispatch(caller)
	endIf
else
	jcp___dispatch(caller)
endIf  ; stringLength(name +type +state +containerName +containerType +value +position +dialogText)

; Indices of 0 and unset are treated equal.
rowIdx = jcp__get(props, "row."+callerName)
if caller != "sayLine" && rowIdx
	var string tblfmt
	rowCount = jcp__get(props, "rows."+callerName)
	colIdx = jcp__get(props, "col."+callerName)
	colCount = jcp__get(props, "cols."+callerName)
	tblfmt = "row %1"
	if rowCount
		tblfmt = tblfmt +" of %2"
	endIf
	if colIdx
		tblfmt = tblfmt +", column %3"
		if colCount
			tblfmt = tblfmt +" of %4"
		endIf
	endIf
	sayUsingVoice(VCTX_Message, formatString(tblfmt, intToString(rowIdx), intToString(rowCount), intToString(colIdx), intToString(colCount)), OT_Position)
endIf  ; rowIdx

if caller != "sayLine" && actionClass != "arrow" && afterField
	say(afterField, OT_Control_Name, True)
endIf
if caller != "sayLine" && actionClass != "arrow"
	; Tutor messages don't speak on SayLine, so beforeTutor doesn't either.
	var
		string beforeTutor
	beforeTutor = jcp__get(props, "beforeTutor")
	if !stringIsBlank(beforeTutor)
		; See JAWS 8's tutorialHelp.jss::sayTutorialHelp() for the rationale for using this SayUsingVoice call.
		; TODO: This may speak in a few undesirable places depending on user settings.
		sayUsingVoice(VCTX_Message, beforeTutor, OT_Line)
	endIf
endIf
return True
endFunction

void function jcp___setCaller(int setting)
if setting
	jcp___key = getCurrentScriptKeyName()
	jcp___caller = getScriptAssignedTo(jcp___key)
	; Clear that shortly.
	scheduleFunction("jcp___setCaller", 5)
else
	jcp___key = ""
	jcp___caller = ""
endIf
endFunction

void function keyPressedEvent(int nKey, string strKeyName, int nIsBrailleKey, int nIsScriptKey)
if nIsBrailleKey || nIsScriptKey
	jcp__reset()
	jcp___setCaller(True)
else
	jcp__reset()
	jcp___setCaller(False)
endIf
keyPressedEvent(nKey, strKeyName, nIsBrailleKey, nIsScriptKey)
endFunction

void function gestureEvent(string gestureName, int IsAssignedToScript)
jcp__reset()
gestureEvent(gestureName, IsAssignedToScript)
endFunction

void function objStateChangedEvent(handle hObj, optional int iObjType, int nChangedState, int nState, int nOldState)
jcp__reset()
objStateChangedEvent(hObj, iObjType, nChangedState, nState, nOldState)
endFunction

void function javaFocusChangedEvent(handle FocusWindow, handle PrevWindow)
jcp__reset()
javaFocusChangedEvent(FocusWindow, PrevWindow)
endFunction

void function valueChangedEvent(handle hwnd, int objId, int childId, int nObjType, string sObjName, string sObjValue,optional int bIsFocusObject)
jcp__reset()
valueChangedEvent(hwnd, objId, childId, nObjType, sObjName, sObjValue, bIsFocusObject)
endFunction

Void Function NameChangedEvent (handle hwnd, int objId, int childId, int nObjType, string sOldName, string sNewName)
jcp__reset()
nameChangedEvent(hwnd, objId, childId, nObjType, sOldName, sNewName)
endFunction

void function descriptionChangedEvent(handle hwnd, int objId, int childId, int nObjType, string sOldDescription, string sNewDescription,optional int bFromFocusObject)
jcp__reset()
descriptionChangedEvent(hwnd, objId, childId, nObjType, sOldDescription, sNewDescription, bFromFocusObject)
endFunction

void function menuModeEvent(handle WinHandle, int mode)
jcp__reset()
menuModeEvent(WinHandle, mode)
endFunction

void function focusPointMovedEvent(int nX, int nY, int nOldX, int nOldY, int nUnit, int nDir, int nTimeElapsed)
; Reset the cache on scripts in case they change focus, values, etc.
jcp__reset()
focusPointMovedEvent(nX, nY, nOldX, nOldY, nUnit, nDir, nTimeElapsed)
endFunction

void function focusChangedEventEx(handle hwndFocus, int nObject, int nChild, handle hwndPrevFocus, int nPrevObject, int nPrevChild, int nChangeDepth)
jcp__reset()
focusChangedEventEx(hwndFocus, nObject, nChild, hwndPrevFocus, nPrevObject, nPrevChild, nChangeDepth)
endFunction

void function sayObjectTypeAndText(int nLevel)
; An override to improve handling of some windows.
var
	collection props,
	handle hwnd,
	string name,
	string buf
if jcp___disabled || jcp___inside || !isPCCursor() || nLevel > 0
	; System disabled, recursive call, non-PC cursor situation, or JAWS 8+ call on non-current control; handle normally.
	sayObjectTypeAndText(nLevel)
	return
endIf

; Look for special handling of the current control.
if jcp__setAll(props, "current", True)
	jcp___inside = True
	jcp__say(props, "sayObjectTypeAndText(" +intToString(nLevel) +")")
	|| sayObjectTypeAndText(nLevel)
	jcp___inside = False
	return
endIf

sayObjectTypeAndText(nLevel)
endFunction

void function sayWindowTypeAndText(handle hwnd)
; Look for special handling of the indicated window.
var
	collection props,
	string sOrigin
if jcp___disabled || jcp___inside
	sayWindowTypeAndText(hwnd)
	return
endIf

if hwnd == getFocus()
	sOrigin = "focus"
elif hwnd == getCurrentWindow()
	sOrigin = "current"
else
	sOrigin = "hwnd"+intToString(hwnd)
endIf
if !jcp___inside && jcp__setAll(props, sOrigin, True)
	jcp___inside = True
	jcp__say(props, "sayWindowTypeAndText(" +intToString(hwnd) +")")
	|| sayWindowTypeAndText(hwnd)
	jcp___inside = False
	return
endIf
sayWindowTypeAndText(hwnd)
endFunction

string function positionInGroup()
if jcp___disabled || jcp___inside
	return positionInGroup()
endIf

var string sOrigin = "current"
var collection props
if !jcp___inside && jcp__setAll(props, sOrigin, True)
	jcp___inside = True
	var string val = jcp__get(props, "position")
	jcp___inside = False
	if val
		return val
	endIf
endIf
return positionInGroup()
endFunction

int function SayByTypeForScriptSayLine()
; Look for special handling of the current line.
var
	collection props,
	string sOrigin
if jcp___disabled || jcp___inside || !isPCCursor()
	return SayByTypeForScriptSayLine()
endIf
sOrigin = "current"
if !jcp__setAll(props, sOrigin, True)
	return SayByTypeForScriptSayLine()
endIf
;var string pn forEach pn in props sayString(formatString("%1 %2", pn, props[pn])) endForEach
jcp___handleMLE(props)
jcp___inside = True
var int result = jcp__say(props, "SayByTypeForScriptSayLine")
|| sayByTypeForScriptSayLine()
jcp___inside = False
return result
endFunction

void function sayLine(int iDrawHighlights, int bDoNotUseObjInfo)
; Look for special handling of the current line.
var
	collection props,
	string sOrigin
if jcp___disabled || jcp___inside || !isPCCursor()
	sayLine(iDrawHighlights, bDoNotUseObjInfo)
	return
endIf
sOrigin = "current"
if !jcp__setAll(props, sOrigin, True)
	sayLine(iDrawHighlights, bDoNotUseObjInfo)
	return
endIf
jcp___handleMLE(props)
jcp___inside = True
jcp__say(props, "sayLine(" +intToString(iDrawHighlights) +", " +intToString(bDoNotUseObjInfo) +")")
|| sayLine(iDrawHighlights, bDoNotUseObjInfo)
jcp___inside = False
endFunction

void function spellLine()
; Look for special handling of the current line.
var
	collection props,
	string sOrigin
if jcp___disabled || jcp___inside || !isPCCursor()
	spellLine()
	return
endIf
sOrigin = "current"
if !jcp__setAll(props, sOrigin, True)
	spellLine()
	return
endIf
jcp___handleMLE(props)
jcp___inside = True
jcp__say(props, "spellLine()")
|| spellLine()
jcp___inside = False
endFunction

int function jcp___handleMLE(collection props)
; Special handling for (and detection of) multiline edit controls.
; Assumes we are dealing with the currently active control.
var handle hwnd = getCurrentWindow()
var int changed = False
var int set_stc = jcp__get(props, "SubtypeCode")
var int isMLE = False
if set_stc && set_stc != WT_Multiline_Edit
	isMLE = False
elif getObjectSubtypeCode() == WT_MultiLine_Edit
; Catches controls that are being specifically set to this type.
|| set_stc == WT_MultiLine_Edit
; Catches controls that JAWS doesn't know are multiline edits.
; The second condition catches read-only multiline edits, which getObjectSubtypeCode() just calls WT_ReadOnlyEdit.
; the SendMessage checks for if the control accepts EM_SetSel, which implies that it's actually an edit control.
; 0x87 is WM_GetDlgCode, and 8 is DLGC_HasSetSel.
|| ((getWindowStyleBits(hwnd) & ES_Multiline)
	&& (sendMessage(hwnd, 0x87, 0, 0) & 8))
		isMLE = True
endIf
var string val = getObjectValue()
if !isMLE
&& (stringContains(val, "\n") || stringContains(val, "\r"))
	isMLE = True
elif !isMLE
&& stringLeft(getWindowClass(hwnd), 11) == "HwndWrapper"
	var object o, int childID
	o = getFocusObject(childID)
	val = o.accValue(childID)
	if (stringContains(val, "\n") || stringContains(val, "\r"))
		isMLE = True
	endIf
endIf
if isMLE
	; Avoid special processing on up/down arrows.
	jcp__set(props, "Mask.arrow", True)
	jcp__set(props, "Mask.sayLine", True)
	; Don't say a control name for SayLine in a multiline edit when it contains something.
	if getObjectValue()
	; or WM_GetTextLength returns non-zero.
	|| (sendMessage(hwnd, 0x0E, 0, 0) > 0)
		jcp__set(props, "Name", " ")
	endIf
	changed = True
endIf
return changed
endFunction

void function sayObjectActiveItem(optional variant sayPositionInfo)
; Look for special handling of the current item.
var
	collection props,
	string sOrigin
var int specifySPI = (sayPositionInfo || stringLength(sayPositionInfo))
if jcp___disabled || jcp___inside
	if specifySPI
		sayObjectActiveItem(sayPositionInfo)
	else
		sayObjectActiveItem()
	endIf
	return
elif getObjectSubtypeCode() == WT_MultiLine_Edit
; The second condition catches read-only multiline edits, which getObjectSubtypeCode() just calls WT_ReadOnlyEdit.
; the SendMessage checks for if the control accepts EM_SetSel, which implies that it's actually an edit control.
; 0x87 is WM_GetDlgCode, and 8 is DLGC_HasSetSel.
|| ((getWindowStyleBits(getCurrentWindow()) & ES_Multiline) && (sendMessage(getCurrentWindow(), 0x87, 0, 0) & 8))
	if specifySPI
		sayObjectActiveItem(sayPositionInfo)
	else
		sayObjectActiveItem()
	endIf
	return
endIf

sOrigin = "current"
if isPCCursor() && jcp__setAll(props, sOrigin, True)
	jcp___inside = True
	if specifySPI
		jcp__say(props, "sayObjectActiveItem(" +intToString(sayPositionInfo) +")")
		|| sayObjectActiveItem(sayPositionInfo)
	else
		jcp__say(props, "sayObjectActiveItem")
		|| sayObjectActiveItem()
	endIf
	jcp___inside = False
	return
endIf
if specifySPI
	sayObjectActiveItem(sayPositionInfo)
else
	sayObjectActiveItem()
endIf
endFunction

int function getTreeViewLevel()
if jcp___disabled || jcp___inside || !isPCCursor()
	return getTreeViewLevel()
endIf
var collection props
if jcp__setAll(props, "current", False)
	; Level 0 explicitly supported here.
	if collectionItemExists(props, "level")
		return stringToInt(jcp__get(props, "Level"))
	endIf
endIf
return getTreeViewLevel()
endFunction

void function sayTreeViewItem()
if jcp___disabled || jcp___inside || !isPCCursor()
	sayTreeViewItem()
	return
endIf
var collection props
if jcp__setAll(props, "current", False)
	sayObjectActiveItem(True)
	return
endIf
sayTreeViewItem()
endFunction

void function say(string txt, int ot, optional int markedUp)
if jcp___disabled || jcp___inside || !isPCCursor()
|| ot != OT_Line
	if markedUp
		say(txt, ot, markedUp)
	else
		say(txt, ot)
	endIf
	return
endIf
var string stk = getScriptCallStack()
if !stringContains(stk, "activeitemchangedevent")
&& !stringContains(stk, "focuschangedevent")
	if markedUp
		say(txt, ot, markedUp)
	else
		say(txt, ot)
	endIf
	return
endIf
var collection props
if jcp__setAll(props, "current", False)
	sayObjectActiveItem(False)
	return
endIf
if markedUp
	say(txt, ot, markedUp)
else
	say(txt, ot)
endIf
endFunction

int function sayTutorialHelp(int iSubType, int isScriptKey)
; Blocks tutor messages when set to blank.
; Also allows iSubtype to be overridden by a subtypeCode property in props.
; Finally allows the subtype to be obtained from a different origin (a specific window).
var
	collection props,
	string msg
if jcp___disabled || jcp___inside || !isPCCursor() || getObjectSubtypeCode() != iSubtype
	; We know nothing of this, so let JAWS figure it out.
	return sayTutorialHelp(iSubType, isScriptKey)
endIf
if jcp__setAll(props, "current", False)
	msg = jcp__get(props, "tutor")
	if stringLength(msg) > 0 && stringIsBlank(msg)
		; Somebody said say nothing, so say nothing.
		return True  ; but say we said what we needed to say.
	elif jcp__getActionClass() == "arrow"
		; No tutor text on arrows.
		return True
	elif !stringIsBlank(msg)
		var int iSpeak = jcp___getTutorialHelpOutputType(isScriptKey)
		sayMessage(iSpeak, msg)
		return True
	endIf
	; For when the subtypeCode is changed.
	var int iSubtype1
	iSubtype1 = jcp__get(props, "subtypeCode")
	; subtypeCode of 0 explicitly allowed here.
	if collectionItemExists(props, "subtypeCode")
		iSubtype = iSubtype1
	else
		; This is just for when jcp__setCustomProps() changes the origin.
		var string sOrigin, handle hwnd
		sOrigin = jcp__get(props, "origin")
		if stringLeft(sOrigin, 4) == "hwnd"
			hwnd = stringToHandle(stringChopLeft(sOrigin, 4))
			saveCursor() invisibleCursor() saveCursor()
			moveToWindow(hwnd)
			iSubtype = getObjectSubtypeCode()
		endIf  ; sOrigin == "hwnd"
	endIf  ; iSubtype1, else
endIf  ; jcp__setAll got something
return sayTutorialHelp(iSubType, isScriptKey)
endFunction

int function jcp___getTutorialHelpOutputType(optional int IsScriptInvoked)
; Private version of same-name function in JAWS' TutorialHelp.jss starting in JAWS 15.
; This version should work back as far as the latest JAWS 11.
If IsScriptInvoked
	if getJFWVersion() >= 1300000
		var int RunningProducts = GetRunningFSProducts()
		; 2 is Product_MAGic; not declared in JAWS 13 or before.
		if RunningProducts == 2
			return OT_HELP
		endIf
	endIf
	;we want an output type which is not ot_tutor,
	;so that it will not be turned off when the user turns off tutor messages.
	;But, we also want an output type which will not flash in braille
	;since we want only speech when the user presses Isnert+Tab.
	return OT_line
EndIf
return OT_tutor
EndFunction

int function sayTutorialHelpHotKey(handle hwnd, int isScriptKey)
var
	collection props,
	string msg
if jcp___disabled || jcp___inside || !isPCCursor() || hwnd != getFocus()
	; We know nothing of this, so let JAWS figure it out.
	return sayTutorialHelpHotKey(hwnd, isScriptKey)
endIf
if jcp__setAll(props, "current", False)
	msg = jcp__get(props, "hotkey")
	if stringLength(msg) > 0
		if stringIsBlank(msg)
			; Somebody said say nothing, so say nothing.
		elif jcp__getActionClass() == "arrow"
			; No tutor text on arrows.
		else
			sayUsingVoice(VCTX_Message, msg, OT_Access_Key)
		endIf
		return True
	endIf
	; This is just for when jcp__setCustomProps() changes the origin.
	var string sOrigin
	sOrigin = jcp__get(props, "origin")
	if stringLeft(sOrigin, 4) == "hwnd"
		hwnd = stringToHandle(stringChopLeft(sOrigin, 4))
		saveCursor() invisibleCursor() saveCursor()
		moveToWindow(hwnd)
	endIf
endIf
sayTutorialHelpHotKey(hwnd, isScriptKey)
endFunction

string function getFrameTutorMessage()
; Called in JAWS 7.1+ to get custom tutor messages.
; Normally this would be getCustomTutorMessage() but that seems not to be called in JAWS 16.
; Cause unknown but this works. [DGL, 2015-09-25]
var
	collection props,
	string msg
msg = ""
if jcp___disabled || jcp___inside || !jcp__setAll(props, "current", False)
	return getFrameTutorMessage()
endIf
msg = jcp__get(props, "customTutor")
if msg
	return msg
endIf
return getFrameTutorMessage()
endFunction

int function jcp___BrailleAddObjectHelper0(string whichProp, int nSubtype)
; Logic for BrailleAddObject* functions to use (wrapped by jcp___BrailleAddObjectHelper though).
var
	string sOrigin,
	collection props,
	int attribs,
	int x, int y,
	string val
if jcp___disabled
	return False
endIf
var int scode = jcp___cachedVal("stc")
if scode && scode != nSubtype
	; BrailleAddObject* function(s) called on a parent dialog control.
	var int level = jcp___findAncestorOfType(nSubtype)
	if level >= 0
		sOrigin = "focus" +intToString(level)
	else
		; TODO:  This may cause some incorrect behavior, but it is hard to know when this should or shouldn't be done.
		; Example:  A .Net edit combo/spin box has been seen to return Spinbox for Braille and Edit Combo for speech.
		; GetObjectSubtypeCode returns ReadOnly Edit at level 0 and Combobox at level 1.
		;sOrigin = "subtype" +intToString(nSubtype)
		sOrigin = "focus"
	endIf
else
	sOrigin = "focus"
endIf
if !jcp__setAll(props, sOrigin, False)
	return False
endIf
if whichProp == "name"
	; Indices of 0 and unset are treated equal.
	val = jcp__get(props, "row.brl")
	if val
		var variant val1
		val1 = jcp__get(props, "col.brl")
		if val1
			val = formatString("r%1c%2", val, val1)
		else
			val = formatString("r%1", val)
		endIf
		BrailleAddString(val, 0,0, 0)
	endIf
	val = jcp__get(props, "beforeField.brl")
	if val
		BrailleAddString(val, 0,0, 0)
	endIf
endIf
val = jcp__get(props, whichProp +".brl")
if !val
	if whichProp == "type"
		scode = jcp__get(props, "subtypeCode.brl")
		; This one is treated the same if 0 or unset.
		if scode
			val = BrailleGetSubtypeString(scode)
			if !val
				val = " "
			endIf
		endIf
	elif whichProp == "state"
		attribs =jcp__get(props, "attributes")
		; Attributes of 0 are supported explicitly here.
		if collectionItemExists(props, "attributes")
			if attribs
				val = BrailleGetStateString (attribs)
			endIf
			if !val
				val = " "
			endIf
		endIf
	endIf
endIf
var int retval = False
if val
	if !stringIsBlank(val)
		x = 0
		y = 0
		; "name" left out deliberately; prevents cursor from being placed on label instead of value in edit fields
		if whichProp == "type" || whichProp == "state" || whichProp == "value"
			x = getCursorCol()
			y = getCursorRow()
		endIf
		;if whichProp == "type" sayString(val) endIf
		BrailleAddString(val, x, y, 0)
	endIf  ; !stringIsBlank(val)
	retval = True
endIf  ; val
if whichProp == "value"
	val = jcp__get(props, "afterField.brl")
	if val
		if !retval
			; No custom value, but put any non-custom one here.
			; The BrailleAddObjectValue in this file should be our caller and so shouldn't run again.
			BrailleAddObjectValue(nSubtype)
		endIf
		BrailleAddString(val, 0,0, 0)
	endIf
endIf
return retval
endFunction

int function jcp___findAncestorOfType(int nSubtype)
; Version of the JAWS 16+ findAncestorOfType function that works in older JAWS versions. Only scans up five levels though. Uses findAncestorOfType() when available.
if getJFWVersion() >= 16000000
	return findAncestorOfType(nSubtype)
endIf
var int level = 1
while level <= 5
	if getObjectSubTypeCode(False, level) == nSubtype
		return level
	endIf
	level = level +1
endWhile
return -1
endFunction

int function jcp___BrailleAddObjectHelper(string whichProp, int nSubtype)
; Called by all BrailleAddObject* functions.
var
	int retval
retval = jcp___BrailleAddObjectHelper0(whichProp, nSubtype)
if retval
	return retval
endIf
return jcp___dispatch("BrailleAddObject" +whichProp +"(" +intToString(nSubtype) +")")
endFunction

; These are all the BrailleAddObject* functions internally recognized as of JAWS 8.0.
int function BrailleAddObjectName(int nSubtype)
return jcp___BrailleAddObjectHelper("Name", nSubtype)
endFunction
int function BrailleAddObjectType(int nSubtype)
return jcp___BrailleAddObjectHelper("Type", nSubtype)
endFunction
int function BrailleAddObjectState(int nSubtype)
return jcp___BrailleAddObjectHelper("State", nSubtype)
endFunction
int function BrailleAddObjectValue(int nSubtype)
return jcp___BrailleAddObjectHelper("Value", nSubtype)
endFunction
int function BrailleAddObjectContainerName(int nSubtype)
return jcp___BrailleAddObjectHelper("ContainerName", nSubtype)
endFunction
int function BrailleAddObjectContainerType(int nSubtype)
return jcp___BrailleAddObjectHelper("ContainerType", nSubtype)
endFunction
int function BrailleAddObjectPosition(int nSubtype)
return jcp___BrailleAddObjectHelper("Position", nSubtype)
endFunction
int function BrailleAddObjectDlgPageName(int nSubtype)
return jcp___BrailleAddObjectHelper("DlgPageName", nSubtype)
endFunction
int function BrailleAddObjectDlgText(int nSubtype)
return jcp___BrailleAddObjectHelper("DlgText", nSubtype)
endFunction
int function BrailleAddObjectContextHelp(int nSubtype)
return jcp___BrailleAddObjectHelper("ContextHelp", nSubtype)
endFunction
int function BrailleAddObjectTime(int nSubtype)
return jcp___BrailleAddObjectHelper("Time", nSubtype)
endFunction
int function BrailleAddObjectLevel(int nSubtype)
return jcp___BrailleAddObjectHelper("Level", nSubtype)
endFunction

int function BrailleCallbackObjectIdentify()
; Allows WT_Unknown to be translated to jcp__get(props, "subtypeCode.brl").
; Set subtypeCode.brl to adjust the Brailled type and the set of BrailleAddObject* functions called for a control.
; Also allows the subtype to be obtained from a different origin (a specific window).
var int tc = getTickCount()
var
	collection props,
	int scode
if !jcp___disabled && !jcp___inside && jcp__setAll(props, "current", False)
	scode = jcp__get(props, "subtypeCode.brl")
	; subtypeCode of 0 explicitly allowed here.
	if collectionItemExists(props, "subtypeCode")
		return scode
	else
		; This is just for when jcp__setCustomProps() changes the origin.
		var string sOrigin, handle hwnd
		sOrigin = jcp__get(props, "origin")
		if stringLeft(sOrigin, 4) == "hwnd"
			hwnd = stringToHandle(stringChopLeft(sOrigin, 4))
			saveCursor() invisibleCursor() saveCursor()
			moveToWindow(hwnd)
			scode = getObjectSubtypeCode()
			return scode
		endIf  ; sOrigin == "hwnd"
	endIf
endIf
return BrailleCallbackObjectIdentify()
endFunction

int function jcp__set(collection props, string whichProp, variant val, /*?*/ int isFunctionCall)
; Set a control property in props.
; Case is not significant in property names.
; If isFunctionCall is True, val is taken to be a function call,
; and jcp__get() will treat it as such. This is implemented
; via an internal .isFunc suffix on property names.
; Special itemPos/Count for setting parts of Position conveniently.
if whichProp == "itemPos" || whichProp == "itemCount"
	var string posdata, int pos
	posdata = jcp__get(props, "Position")
	if !posdata
		posdata = positionInGroup()
		if !posdata
			posdata = jcp___MOfN
		endIf
	endIf
	if whichProp == "itemPos"
		pos = stringContains(posdata, " ")
		posdata = stringChopLeft(" " +intToString(val) +stringChopLeft(posdata, pos-1), 1)
	else
		pos = stringContainsFromRight(posdata, " ")
		posdata = stringChopLeft(" " +stringLeft(posdata, pos) +intToString(val), 1)
	endIf
	whichProp = "position"
	val = jcp___vcast(posdata)
endIf
if isFunctionCall
	props[whichProp+".isFunc"] = 1
endIf
props[whichProp] = val
return True
endFunction

void function jcp__reportProps(collection props, optional int virtualize)
; Report properties, virtualizing if requested.
var string k, string v, string buf
forEach k in props
	v = props[k]
	buf = formatString("%1: %2", k, v)
	if virtualize
		sayMessage(OT_User_Buffer, buf)
	else
		sayUsingVoice(VCTX_Message, buf, OT_Message)
	endIf
endForEach
endFunction

;======================== End module jcp (from jcpdict) ========================

